import unittest
from datetime import timedelta

import numpy as np
from numpy.testing import assert_allclose
from skyfield.api import load as skyf_load
from skyfield.iokit import parse_tle_file

from sat_geo_solver.observe import (Observe, distance_to, doppler,
                                    light_seconds, range_and_rate,
                                    relativistic_doppler)


class TestObserve(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Load timescale
        cls.ts = skyf_load.timescale()

        # Load satellite from g19.tle
        with open('tests/g19.tle', 'rb') as f:
            cls.satellites = list(parse_tle_file(f, cls.ts))
        cls.sat = cls.satellites[0]

        # Observer location from notebook
        cls.cos_ll = [+38.4, -104.82]

        # Datetimes from notebook (±12h from epoch, 15m steps)
        epoch_dt = cls.sat.epoch.utc_datetime()
        start_dt = epoch_dt - timedelta(hours=12)
        end_dt = epoch_dt + timedelta(hours=12)

        cls.dt_observables = []
        curr_dt = start_dt
        while curr_dt <= end_dt:
            cls.dt_observables.append(curr_dt)
            curr_dt += timedelta(minutes=15)

        # Observe instance
        cls.observe = Observe(cls.sat, cls.dt_observables)
        cls.observe_epoch = Observe(cls.sat, cls.sat.epoch.utc_datetime())

    def test_distance_to_single_epoch(self):
        expected = distance_to(self.sat, self.cos_ll, self.sat.epoch.utc_datetime())
        actual = self.observe_epoch.distance_to(self.cos_ll)
        assert_allclose(actual, expected)

    def test_distance_to(self):
        # Compare Observe.distance_to with standalone distance_to
        expected = distance_to(self.sat, self.cos_ll, self.dt_observables)
        actual = self.observe.distance_to(self.cos_ll)

        assert_allclose(actual, expected)

    def test_range_and_rate_single_epoch(self):
        expected_range, expected_rate = range_and_rate(self.sat, self.cos_ll, self.sat.epoch.utc_datetime())
        actual_range, actual_rate = self.observe_epoch.range_and_rate(self.cos_ll)

        assert_allclose(actual_range, expected_range)
        assert_allclose(actual_rate, expected_rate)

    def test_range_and_rate(self):
        # Compare Observe.range_and_rate with standalone range_and_rate
        expected_range, expected_rate = range_and_rate(self.sat, self.cos_ll, self.dt_observables)
        actual_range, actual_rate = self.observe.range_and_rate(self.cos_ll)

        assert_allclose(actual_range, expected_range)
        assert_allclose(actual_rate, expected_rate)

    def test_light_seconds_single_epoch(self):
        expected = light_seconds(self.sat, self.cos_ll, self.sat.epoch.utc_datetime())
        actual = self.observe_epoch.light_seconds(self.cos_ll)
        assert_allclose(actual, expected)

    def test_light_seconds(self):
        # Compare Observe.light_seconds with standalone light_seconds
        expected = light_seconds(self.sat, self.cos_ll, self.dt_observables)
        actual = self.observe.light_seconds(self.cos_ll)

        assert_allclose(actual, expected)

    def test_pos_vel_xyz_output(self):
        # Verify that pos_xyz and vel_xyz return numpy arrays with xyz
        pos = self.observe.pos_xyz()
        vel = self.observe.vel_xyz()

        self.assertIsInstance(pos, np.ndarray)
        self.assertIsInstance(vel, np.ndarray)

        # Since it's a list of datetimes, shape should be (3, N) for skyfield position/velocity
        self.assertEqual(pos.shape, (3, len(self.dt_observables)))
        self.assertEqual(vel.shape, (3, len(self.dt_observables)))


def test_relativistic_doppler_against_doppler():
    freq = 14.25e9
    vel = 10
    r_dop = relativistic_doppler(freq, vel)
    assert_allclose(r_dop, doppler(freq, vel))


if __name__ == '__main__':
    unittest.main()



from datetime import datetime
from typing import List, Tuple

import numpy as np
from skyfield.api import load as skyf_load
from skyfield.api import wgs84
from skyfield.constants import C as SPEED_OF_LIGHT
from skyfield.timelib import Time

__ts__ = skyf_load.timescale()


def doppler(frequency: float | np.ndarray, range_rate: float | np.ndarray) -> float | np.ndarray:
    """
    Classical (non-relativistic) Doppler: compute received frequency from a line-of-sight range rate.

    Sign convention (Skyfield):
      - range_rate > 0 means range increasing (receding) -> received frequency decreases

    :param frequency: Transmitted frequency in Hz
    :param range_rate: Line-of-sight range rate in m/s (positive = moving away)
    :return: Received frequency in Hz
    """
    return frequency * (1.0 - (range_rate / SPEED_OF_LIGHT))


def relativistic_doppler(frequency: float | np.ndarray, range_rate: float | np.ndarray) -> float | np.ndarray:
    """
    Relativistic Doppler: compute received frequency from a line-of-sight range rate.

    Sign convention (Skyfield):
      - range_rate > 0 means range increasing (receding)
      - Doppler beta is positive for approaching, so beta = -range_rate / c

    :param frequency: Transmitted frequency in Hz
    :param range_rate: Line-of-sight range rate in m/s (positive = moving away)
    :return: Received frequency in Hz
    """
    beta = -range_rate / SPEED_OF_LIGHT
    return frequency * np.sqrt((1.0 + beta) / (1.0 - beta))


def dt_to_ts(dt: datetime | List[datetime]) -> Time:
    """
    Convert datetime or list of datetimes to Skyfield Time.

    :param dt: Datetime or list of datetimes
    :return: Skyfield Time object
    """
    if isinstance(dt, datetime):
        return __ts__.from_datetime(dt)
    elif isinstance(dt, list):
        return __ts__.from_datetimes(dt)
    else:
        raise TypeError(f"Invalid type for 'dt': {type(dt)}")


def distance_to(sat, from_pos: list | np.ndarray, at: datetime | List[datetime]) -> float | List[float]:
    """
    Provide the distance to the satellite from a given position at a given time.

    :param sat: Satellite to observe
    :param from_pos: Position to observe from
    :param at: Time to observe at
    :return: Distance in meters
    """
    pos = wgs84.latlon(*from_pos)
    loc_diff = sat - pos
    observe_ts = dt_to_ts(at)
    return loc_diff.at(observe_ts).distance().m


def range_and_rate(sat, from_pos: list | np.ndarray, at: datetime | List[datetime]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculate the range and rate of a satellite from a given position at a given time.

    :param sat: Satellite to observe
    :param from_pos: Position to observe from
    :param at: Time to observe at
    :return: Tuple of range and rate in meters and meters per second
    """
    pos = wgs84.latlon(*from_pos)
    loc_diff = sat - pos
    observe_ts = dt_to_ts(at)
    _, _, range_distance, _, _, range_rate = loc_diff.at(observe_ts).frame_latlon_and_rates(pos)
    return range_distance.m, range_rate.m_per_s


def light_seconds(sat, from_pos: list | np.ndarray, at: datetime | List[datetime]) -> float | np.ndarray:
    """
    Calculate the light time between the observer and the satellite.

    :param sat: Satellite to observe
    :param from_pos: Position to calculate light time from
    :param at: Time to calculate light time at
    :return: Light time in seconds
    """
    pos = wgs84.latlon(*from_pos)
    loc_diff = sat - pos
    observe_ts = dt_to_ts(at)
    return loc_diff.at(observe_ts).distance().light_seconds()


class Observe:

    def __init__(self, sat, at: datetime | List[datetime]):
        self.sat = sat
        self.observe_ts = dt_to_ts(at)
        self.sat_at = sat.at(self.observe_ts)

    def distance_to(self, from_pos: list | np.ndarray) -> float | List[float]:
        """
        Provide the distance to the satellite from a given position.

        :param from_pos: Position to observe from
        :return: Distance to the satellite in meters
        """
        loc_diff = self.sat_at - wgs84.latlon(*from_pos).at(self.observe_ts)
        return loc_diff.distance().m

    def range_and_rate(self, from_pos: list | np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculate the range and rate of the satellite from a given position.

        :param from_pos: Position to observe from
        :return: Range distance in meters, range rate in meters per second
        """
        pos = wgs84.latlon(*from_pos)
        loc_diff = self.sat_at - pos.at(self.observe_ts)
        _, _, range_distance, _, _, range_rate = loc_diff.frame_latlon_and_rates(pos)
        return range_distance.m, range_rate.m_per_s

    def pos_xyz(self) -> np.ndarray:
        """
        Produce the xyz position of the satellite in meters.

        :return: np.ndarray
        """
        return self.sat_at.position.m

    def vel_xyz(self) -> np.ndarray:
        """
        Produce the xyz velocity of the satellite in meters per second.

        :return: np.ndarray
        """
        return self.sat_at.velocity.m_per_s

    def light_seconds(self, from_pos: list | np.ndarray) -> float | np.ndarray:
        """
        Calculate the light time between the observer and the satellite.

        :param from_pos: Position to calculate light time from
        :return: Light time in seconds
        """
        pos = wgs84.latlon(*from_pos)
        loc_diff = self.sat_at - pos.at(self.observe_ts)
        return loc_diff.distance().light_seconds()

    def downlink_received_frequency(self,
                                    from_pos: list | np.ndarray,
                                    to_pos: list | np.ndarray,
                                    uplink: float,
                                    translation_frequency: float) -> float | np.ndarray:
        """
        Bent-pipe link budget in terms of frequency only.

        Steps:
          1) Uplink: ground transmitter -> satellite receiver (uplink Doppler)
          2) Translate at satellite: f_down_tx = f_up_rx - translation_frequency
          3) Downlink: satellite transmitter -> ground receiver (downlink Doppler)

        :param from_pos: Position from which the signal is emitted.
        :param to_pos: Position at which the signal is received.
        :param uplink: Uplink frequency in Hz.
        :param translation_frequency: Translation frequency in Hz.
        :return: Final received downlink frequency at the ground receiver in Hz
        """

        _, rr_uplink = self.range_and_rate(from_pos)
        f_up_rx_at_satellite = relativistic_doppler(uplink, rr_uplink)

        f_down_tx = f_up_rx_at_satellite - translation_frequency

        _, rr_downlink = self.range_and_rate(to_pos)
        return relativistic_doppler(f_down_tx, rr_downlink)

    def doppler_shift(self,
                      from_pos: list | np.ndarray,
                      to_pos: list | np.ndarray,
                      uplink: float,
                      translation_frequency: float) -> float | np.ndarray:
        """
        Final bent-pipe Doppler shift in Hz, relative to the nominal translated frequency:
            f_nominal = uplink - translation_frequency

        :param from_pos: Position from which the signal is emitted.
        :param to_pos: Position at which the signal is received.
        :param uplink: Uplink frequency in Hz.
        :param translation_frequency: Translation frequency in Hz.
        :return: (f_down_rx_at_ground - f_nominal) in Hz
        """
        f_down_rx_at_ground = self.downlink_received_frequency(
            from_pos=from_pos,
            to_pos=to_pos,
            uplink=uplink,
            translation_frequency=translation_frequency,
        )
        f_nominal = uplink - translation_frequency
        return f_down_rx_at_ground - f_nominal

    def fdoa(self,
             from_pos: list | np.ndarray,
             rx1_pos: list | np.ndarray,
             rx2_pos: list | np.ndarray,
             uplink: float,
             translation_frequency: float) -> float | np.ndarray:
        """
        Frequency Difference of Arrival (FDOA) for a bent-pipe satellite:
            FDOA = f_rx(rx1) - f_rx(rx2)

        :param from_pos: Position from which the signal is emitted.
        :param rx1_pos: Position of the first receiver.
        :param rx2_pos: Position of the second receiver.
        :param uplink: Uplink frequency in Hz.
        :param translation_frequency: Translation frequency in Hz.
        :return: Frequency difference in Hz (same shape as the underlying observation times)
        """
        f1 = self.downlink_received_frequency(
            from_pos=from_pos,
            to_pos=rx1_pos,
            uplink=uplink,
            translation_frequency=translation_frequency,
        )
        f2 = self.downlink_received_frequency(
            from_pos=from_pos,
            to_pos=rx2_pos,
            uplink=uplink,
            translation_frequency=translation_frequency,
        )
        return f1 - f2

    def __repr__(self):
        return f"Observe(sat={self.sat}, at={self.observe_ts})"



from datetime import datetime
from typing import List

import numpy as np

from sat_geo_solver.observe import Observe


class TwoSat:

    def __init__(self, primary_satellite,
                 secondary_satellite,
                 at: datetime | List[datetime],
                 primary_receiver: list | np.ndarray,
                 secondary_receiver: list | np.ndarray,):
        self.primary = Observe(primary_satellite, at)
        self.secondary = Observe(secondary_satellite, at)
        self.at = at
        self.primary_receiver = primary_receiver
        self.secondary_receiver = secondary_receiver

    def dto(self, from_pos: list | np.ndarray) -> float | np.ndarray:
        """
        Calculate the differential time offset of a signal from a given position between the two satellites.
        Secondary - Primary

        :param from_pos: Position from which the signal is emitted. [lat_deg, lon_deg, elevation_m]
        :return: Time difference of arrival in light seconds
        """
        primary_signal_path = (
            self.primary.light_seconds(from_pos)
            + self.primary.light_seconds(self.primary_receiver)
        )
        secondary_signal_path = (
            self.secondary.light_seconds(from_pos)
            + self.secondary.light_seconds(self.secondary_receiver)
        )
        return secondary_signal_path - primary_signal_path

    def dfo(self,
            from_pos: list | np.ndarray,
            uplink: float,
            translation_frequency: float) -> float | np.ndarray:
        """
        Calculate the differential frequency offset between the two satellite paths for a bent-pipe link.
        Secondary - Primary

        This models the case where the same uplink is (primarily) received by the primary satellite,
        but some energy is also received by the secondary satellite (e.g., sidelobes), and each satellite
        downlinks to its own receiver.

        :param from_pos: Position from which the signal is emitted. [lat_deg, lon_deg, elevation_m]
        :param uplink: Uplink frequency in Hz.
        :param translation_frequency: Translation frequency in Hz.
        :return: Differential received frequency in Hz (secondary_rx - primary_rx)
        """
        f_primary = self.primary.downlink_received_frequency(
            from_pos=from_pos,
            to_pos=self.primary_receiver,
            uplink=uplink,
            translation_frequency=translation_frequency,
        )
        f_secondary = self.secondary.downlink_received_frequency(
            from_pos=from_pos,
            to_pos=self.secondary_receiver,
            uplink=uplink,
            translation_frequency=translation_frequency,
        )
        return f_secondary - f_primary